/*
 *    Copyright  1996, 1997, 1998, 1999, 2000 by Remco Treffkorn
 *    Distributed under the terms of the Q Public License. Please see
 *    the included file LICENSE.
 */
/******************************************************************************
 *  projtool.c   projection and parameterization routines                     *
 ******************************************************************************/

#include "vector.h"
#include "units.h"
#include <math.h>

#undef TESTPROJ
#define TO_RADIANS PI/351562

void LatLongToXY(MapStruct * ms, long lat, long lon, int *x, int *y);

void Project(ms, lat, lon, x, y)
MapStruct *ms;
long lat, lon;
int *x, *y;
{
    long xm;
    long xp;

    if (ms->proj) LatLongToXY(ms, lat, lon, x, y);
    else if (ms->HalfWidU) {
#ifdef TESTPROJ
	*x = - (MU_TO_PIX(ms, ms->CtrLonU - SubMod(lon, ms->CtrLonU))) - ms->WidthP;
	*y = MU_TO_PIX(ms, ms->CtrLatU - lat) + ms->HeightP / 2;
#else

# if 0
	xp = cos((double)(abs(lat)>>9) * PI/(DEG180_LU >> 9)) * (1 << XPU_SHIFT);
	xm = LONU_TO_MU(SubMod(lon,ms->CtrLonU),xp,ms->GenShift);
	*x = MU_TO_PIX(ms, ms->HalfWidU + xm);
	*y = MU_TO_PIX(ms, ms->CtrLatU - lat) + ms->HeightP/2;
# else
	xp = cos((double)(abs(lat)>>9) * TO_RADIANS) * 2048;
	xm = (((SubMod(lon,ms->CtrLonU))>>(ms->GenShift))*(float)(xp))/
		pow(2,XPU_SHIFT-ms->GenShift);
	*x = MU_TO_PIX(ms, ms->HalfWidU + xm);
	*y = MU_TO_PIX(ms, ms->CtrLatU - lat) + ms->HeightP/2;
# endif

#endif
    }
}

void nProject(ms, lat, lon, x, y)	/* same as above, with lon pre-normalized on CtrLonU */
MapStruct *ms;
long lat, lon;
int *x, *y;
{
    long xm;
    long xp;

    if (ms->proj) LatLongToXY(ms, lat, lon + ms->CtrLonU, x, y);
    else {
#ifdef TESTPROJ
	*x = - (MU_TO_PIX(ms, ms->CtrLonU - lon)) - ms->WidthP;
	*y = MU_TO_PIX(ms, ms->CtrLatU - lat) + ms->HeightP / 2;
#else

# if 0
	xp = cos((double)(abs(lat)>>9) * PI/(DEG180_LU >> 9)) * (1 << XPU_SHIFT);
	xm = LONU_TO_MU(lon,xp,ms->GenShift);
	*x = MU_TO_PIX(ms, ms->HalfWidU + xm);
	*y = MU_TO_PIX(ms, ms->CtrLatU - lat) + ms->HeightP/2;
# else
	xp = cos((double)(abs(lat)>>9) * TO_RADIANS) * 2048;
	xm = (((lon)>>(ms->GenShift))*(float)(xp))/pow(2,XPU_SHIFT-ms->GenShift);
	*x = MU_TO_PIX(ms, ms->HalfWidU + xm);
	*y = MU_TO_PIX(ms, ms->CtrLatU - lat) + ms->HeightP/2;
# endif

#endif
    }
}

/***** Determine scale and other parameters from requested parameters *****/

static double imageRatio, mapRatio;

void map_params(MapStruct *ms)
{
    long HalfHtU, BigLatU;
    long WidthU, HeightU, WidthLU;
    unsigned short xp;

    if ((ms->minWidthD >= 32) || (ms->HeightD >= 16))
	ms->GenShift = GEN_SHIFT << 1;
    else if ((ms->minWidthD >= 1) || (ms->HeightD >= .5))
	ms->GenShift = GEN_SHIFT;
    else
	ms->GenShift = 0;

    ms->CtrLatU = DEG_TO_LU(ms->CtrLatD);
    ms->CtrLonU = DEG_TO_LU(ms->CtrLonD);

    HeightU = DEG_TO_LU(ms->HeightD);
    HalfHtU = HeightU >> 1;

    if (abs(ms->CtrLatU) < HalfHtU)
	BigLatU = 0;
    else
	BigLatU = ms->CtrLatU + (ms->CtrLatD < 0 ? HalfHtU : -(HalfHtU));
    xp = cos((double)(abs(BigLatU)>>9) * PI/(DEG180_LU >> 9)) * (1 << XPU_SHIFT);

    WidthU = LONU_TO_MU(DEG_TO_LU(ms->minWidthD), xp, ms->GenShift);

    /* fit map to image and determine scale */
    mapRatio = (float) HeightU / WidthU;
    imageRatio = (float) ms->HeightP / ms->WidthP;

    if (imageRatio >= mapRatio) {	/* image narrower than map--fit by width */
	HeightU = WidthU * imageRatio;
	ms->HeightD = LU_TO_DEG(HeightU);
	ms->muPerPix = WidthU / ms->WidthP;
    } else {			/* image wider than map--fit by height */
	WidthU = HeightU / imageRatio;
	ms->minWidthD = LU_TO_DEG(MU_TO_LONU(WidthU, xp, ms->GenShift));
	ms->muPerPix = HeightU / ms->HeightP;
    }

    ms->scale = ms->muPerPix * AVGPPI * IN_PER_MU;

    /* determine maximum width at wings of map */
    ms->HalfWidU = WidthU >> 1;
    HalfHtU = (HeightU >> 1) + (8 * ms->muPerPix);	// last bit is an 8-pixel buffer
    if (abs(ms->CtrLatU) > DEG90_LU - HalfHtU) {
	WidthLU = DEG360_LU - 1;
    } else {
	BigLatU = ms->CtrLatU + (ms->CtrLatD > 0 ? HalfHtU : -(HalfHtU));
	xp = cos((double)(abs(BigLatU)>>9) * PI/(DEG180_LU >> 9)) * (1 << XPU_SHIFT);
	WidthLU = MU_TO_LONU(WidthU, xp, ms->GenShift);
	if (WidthLU >= DEG360_LU)
	    WidthLU = DEG360_LU - 1;
    }
    ms->maxWidthD = LU_TO_DEG(WidthLU);
    if (ms->maxWidthD > 32)
	ms->GenShift = GEN_SHIFT << 1;
    else if (ms->maxWidthD > 1)
	ms->GenShift = GEN_SHIFT;
    else
	ms->GenShift = 0;

    /* define normalized bounding box */
    ms->nBoundsU.SLat = ms->CtrLatU - HalfHtU;
    if (ms->nBoundsU.SLat < -DEG90_LU)
	ms->nBoundsU.SLat = -DEG90_LU;
    ms->nBoundsU.NLat = ms->CtrLatU + HalfHtU;
    if (ms->nBoundsU.NLat > DEG90_LU)
	ms->nBoundsU.NLat = DEG90_LU;
    ms->nBoundsU.ELon = WidthLU / 2;
    ms->nBoundsU.WLon = -WidthLU / 2;

    /* determine map edges */
    ms->BoundsU.SLat = ms->nBoundsU.SLat;
    ms->EdgeSD = LU_TO_DEG(ms->nBoundsU.SLat);
    ms->BoundsU.NLat = ms->nBoundsU.NLat;
    ms->EdgeND = LU_TO_DEG(ms->nBoundsU.NLat);
    ms->BoundsU.ELon = ms->CtrLonU + ms->nBoundsU.ELon;
    if (ms->BoundsU.ELon > DEG180_LU)
	ms->BoundsU.ELon -= DEG360_LU;
    ms->EdgeED = LU_TO_DEG(ms->BoundsU.ELon);
    ms->BoundsU.WLon = ms->CtrLonU + ms->nBoundsU.WLon;
    if (ms->BoundsU.WLon < -DEG180_LU)
	ms->BoundsU.WLon += DEG360_LU;
    ms->EdgeWD = LU_TO_DEG(ms->BoundsU.WLon);

//    report(ms);
}


float MercatorLatToY(float lat)
{
    return log(tan(0.5 * (PIOVER2 + lat)));
}


float MercatorLongToX(float lambda0, float longitude)
{
    float x;

    x = longitude - lambda0;

    if (x > PI)
	x -= TWOPI;

    if (x < -PI)
	x += TWOPI;

    return x;
}



int mapPointIsGood;

void LatLongToXY(MapStruct * ms, long lat, long lon, int *x, int *y)
{
    float lml0, cosphi, sinphi, cosc, coslml0, cc;
    float latitude, longitude;

    latitude = LU_TO_DEG(lat) * DEG2RAD;
    longitude = LU_TO_DEG(lon) * DEG2RAD;

    switch (ms->proj) {

    default:
    case MERCATOR:
	*x = (ms->WidthP/2) + (MercatorLongToX(ms->CtrLonD * DEG2RAD, longitude) * ms->WidthP);
	*y = ms->HeightP - MercatorLatToY(latitude) * ms->HeightP;
	break;

#if 0
    case ORTHOGRAPHIC:
	lml0 = longitude - ms->lambda0;
	cosphi = cos(latitude);
	sinphi = sin(latitude);
	coslml0 = cos(lml0);
	cc = cosphi * coslml0;

	*x = (cosphi * sin(lml0)) * ms->radius;
	*y = ms->rcosphi1 * sinphi - ms->rsinphi1 * cc;

	/* see if it is a good point */

	cosc = ms->sinphi1 * sinphi + mdata->cosphi1 * cc;

	mapPointIsGood = (cosc > 0.001);
	break;
#endif
    }

}

#if 0
report(MapStruct * ms)
{
    fprintf(stderr,"\n/* parameters specified by user */\n");
    fprintf(stderr,"int GridSpace;                        %d\n", ms->GridSpace);
    fprintf(stderr,"float CtrLonD, CtrLatD;               %f  %f\n", ms->CtrLonD, ms->CtrLatD);
    fprintf(stderr,"float minWidthD, maxWidthD, HeightD;  %f  %f  %f\n",ms->minWidthD,ms->maxWidthD,ms->HeightD);
    fprintf(stderr,"int WidthP, HeightP;                  %d  %d\n", ms->WidthP, ms->HeightP);
    fprintf(stderr,"/* calculated parameters */\n");
    fprintf(stderr,"long CtrLatU, CtrLonU;                %ld  %ld\n", ms->CtrLatU, ms->CtrLonU);
    fprintf(stderr,"long HalfWidU;                        %ld\n", ms->HalfWidU);
    fprintf(stderr,"MBR BoundsU, nBoundsU;\n"); 
    fprintf(stderr,"    Bounds: Lon=%i %i, Lat=%i %i\n",ms->BoundsU.WLon,ms->BoundsU.ELon,
                                                ms->BoundsU.SLat,ms->BoundsU.NLat);/* */
    fprintf(stderr,"   nBounds: Lon=%i %i, Lat=%i %i\n",ms->nBoundsU.WLon,ms->nBoundsU.ELon,
                                                 ms->nBoundsU.SLat,ms->nBoundsU.NLat);/* */
    fprintf(stderr,"float EdgeSD, EdgeND;                 %f  %f\n", ms->EdgeSD, ms->EdgeND);
    fprintf(stderr,"float EdgeWD, EdgeED;                 %f  %f\n", ms->EdgeWD, ms->EdgeED);
    fprintf(stderr,"long muPerPix, scale;                 %ld  %ld\n", ms->muPerPix, ms->scale);
    fprintf(stderr,"int GenShift;                         %d\n\n", ms->GenShift);
}
#endif
